﻿package  net.nicoptere.marchingcubes
{
	
	/**
	 * @author nicoptere
	 * http://en.nicoptere.net/
	 */
	public class Dataset
	{
		
		public var data:Array;
		
		public var minVal:Number;
		public var maxVal:Number;
		
		public var isoLevel:Number;
		
		public var x:Number;
		public var y:Number;
		public var z:Number; 
		
		/**
		 * instantiates a new Dataset to be prcessed by the Marchincube algorithm
		 * @param	data a 3D array of scalar values
		 * @param	minVal the lowest value of the set
		 * @param	maxVal the highest value of the set
		 * @param	isoLevel the threshold
		 * @param	x width of the data array
		 * @param	y height of the data array
		 * @param	z depth of the data array
		 */
		public function Dataset(data:Array = null, minVal:Number = 0, maxVal:Number = 1, isoLevel:Number = 0, x:Number = 1, y:Number = 1, z:Number = 1 ) 
		{ 
			
			this.data = data;
			this.minVal = minVal;
			this.maxVal = maxVal;
			this.isoLevel = isoLevel;
			this.x = x;
			this.y = y;
			this.z = z;
			
		};
		
		/**
		 * dataset helper: creates a random scalar field based on metaballs
		 * @param	_x : width
		 * @param	_y : height
		 * @param	_z : depth
		 */
		public function randomShape( _x:int = 1, _y:int = 1, _z:int = 1 ):void
		{
			data = [];
			
			x = _x;
			y = _y;
			z = _z;
			
			minVal = Number.MAX_VALUE;
			maxVal = Number.MIN_VALUE;
			
			var i:int;
			var j:int;
			var k:int;
			
			var dx:Number;
			var dy:Number;
			var dz:Number;
			
			var rx:Number = x / 2;
			var ry:Number = y / 2;
			var rz:Number = z / 2;
			
			var sp:Object;
			var spheres:Array = [];
			
			i = 20;
			while ( i-- )
			{
				dx = int( rx ) + int( (Math.random() - .5) * x );
				dy = int( ry ) + int( (Math.random() - .5) * y );
				dz = int( rz ) + int( (Math.random() - .5) * z );
				
				sp = { x:dx, y:dy, z:dz, r: x/2 };
				
				spheres.push( sp );
				
			}
			
			for (i = 0; i <= x; i++) 
			{
				data.push( [] );
				for (j = 0; j <= y; j++) 
				{
					data[ i ].push( [] );
					for (k = 0; k <= z; k++) 
					{
						
						data[ i ][ j ].push( 128 );
						
						for each( sp in spheres )
						{
							
							dx = i - sp.x;
							dy = j - sp.y;
							dz = k - sp.z;
							
							var dist:Number = Math.sqrt( dx * dx + dy * dy + dz * dz );
							if ( dist < sp.r ) 
							{
								data[ i ][ j ][ k ] +=  dist;
							}
							
						}
						if ( data[ i ][ j ][ k ] < minVal ) minVal = data[ i ][ j ][ k ];
						if ( data[ i ][ j ][ k ] > maxVal ) maxVal = data[ i ][ j ][ k ];
					}
				}
			}
			isoLevel = minVal + ( maxVal - minVal ) / 2;
			return;
		}
		
		/**
		 * dataset helper: creates a sphere fitting the given dataset size
		 * @param	_x : width
		 * @param	_y : height
		 * @param	_z : depth
		 */
		public function generateSphere( _x:int = 1, _y:int = 1, _z:int = 1 ):void
		{
			data = [];
			
			x = ( _x % 2 == 0 ) ? ( _x + 1 ) : _x;
			y = ( _y % 2 == 0 ) ? ( _y + 1 ) : _y;
			z = ( _z % 2 == 0 ) ? ( _z + 1 ) : _z;
			
			minVal = Number.MAX_VALUE;
			maxVal = Number.MIN_VALUE;
			
			var i:int;
			var j:int;
			var k:int;
			
			
			var dx:Number;
			var dy:Number;
			var dz:Number;
			
			var rx:Number = x / 2;
			var ry:Number = y / 2;
			var rz:Number = z / 2;
			
			var val:Number;
			
			for (i = 0; i <= x; i++) 
			{
				data.push( [] );
				
				for (j = 0; j <= y; j++) 
				{
					data[ i ].push( [] );
					
					for (k = 0; k <= z; k++) 
					{
						
						data[ i ][ j ].push( 0 );
						
						dx = rx - i;
						dy = ry - j;
						dz = rz - k;
						
						val = ( ( 1 - ( ( dx * dx ) / (rx * rx ) ) ) + ( 1 - ( ( dy * dy ) / (ry * ry ) ) ) + ( 1 - ( ( dz * dz ) / (rz * rz ) ) ) ) * 256;
						data[ i ][ j ][ k ] +=  val;
						
						if ( data[ i ][ j ][ k ] < minVal ) minVal = data[ i ][ j ][ k ];
						if ( data[ i ][ j ][ k ] > maxVal ) maxVal = data[ i ][ j ][ k ];
					}
				}
			}
			isoLevel = int(  maxVal * Math.PI / 4 );
		}
		
		/**
		 * dataset helper: creates a random scalar field with random values
		 * @param	_x : width
		 * @param	_y : height
		 * @param	_z : depth
		 */
		public function noise( _x:int = 1, _y:int = 1, _z:int = 1 ):void
		{
			data = [];
			
			x = _x;
			y = _y;
			z = _z;
			
			minVal = Number.MAX_VALUE;
			maxVal = Number.MIN_VALUE;
			
			var i:int;
			var j:int;
			var k:int;
			
			for (i = 0; i <= x; i++) 
			{
				data.push( [] );
				for (j = 0; j <= y; j++) 
				{
					data[ i ].push( [] );
					for (k = 0; k <= z; k++) 
					{
						
						data[ i ][ j ].push( Math.random()*256 );
						if ( data[ i ][ j ][ k ] < minVal ) minVal = data[ i ][ j ][ k ];
						if ( data[ i ][ j ][ k ] > maxVal ) maxVal = data[ i ][ j ][ k ];
					}
				}
			}
			isoLevel = maxVal / 2;
			return;
		}
		
		/**
		 * dataset helper: creates a noisy surface pretty much like a terrain
		 * @param	_x : width
		 * @param	_y : height
		 * @param	_z : depth
		 */
		public function seedStep( _x:int = 1, _y:int = 1, _z:int = 1 ):void
		{
			data = [];
			
			x = _x;
			y = _y;
			z = _z;
			
			minVal = Number.MAX_VALUE;
			maxVal = Number.MIN_VALUE;
			
			var i:int;
			var j:int;
			var k:int;
			
			var seed:Number = 0;
			var step:Number = 32;
			
			for (i = 0; i <=x; i++) 
			{
				data.push( [] );
				for (j = 0; j <= y; j++) 
				{
					data[ i ].push( [] );
					for (k = 0; k <= z; k++) 
					{
						
						data[ i ][ j ].push( 0 );
						data[ i ][ j ][ k ] = seed + Math.random() * step;
						
						if ( data[ i ][ j ][ k ] < minVal ) minVal = data[ i ][ j ][ k ];
						if ( data[ i ][ j ][ k ] > maxVal ) maxVal = data[ i ][ j ][ k ];
					}
				}
				seed += step;
			}
			isoLevel = maxVal / 2;
			return;
		}
		
		public function toString( ):String 
		{
			return '[Dataset( data:'+data.toString()+', minVal:'+minVal+', maxVal:'+maxVal+', isoLevel:'+isoLevel+', x:'+y+', y:'+y+', z:'+z+' )]';  
		}
	}
}